from math import log10

# a player that never lost a game will have elo score of 1e30 and elo_rating of 9999
# a player that never win a game will have elo score of 1e-30 and elo_rating of 0
HIGHEST_ELO_SCORE = 1e30
LOWEST_ELO_SCORE = 1e-30
HIGHEST_ELO_RATING = 9999
LOWEST_ELO_RATING = 0

# Upperbound estimate of delta ELO_RATING per delta ELO_SCORE:
# If ELO change from 2000 to 2000.1, the score_ratio=ln10(elo/400), and it will change by 2e-5
# so the delta(ELO)/delta(score_ratio) is about 5e4. Making it 1e6 to be safe.
MAX_DELTA_ELO_RATING_PER_DELTA_SCORE_RATIO = 1e6


def get_mle_elo_scores(initial_elo_scores, wins_dict, elo_rating_error_bar=0.1):
    """
    Calculate elo_scores using MLE algorithm.
    MLE refers to maximum likelyhood estimation. The elo_score here is proportional to win odd, and
      elo_rating = elo_base + elo_scale * log10(elo_score)

    Let's say there are two models, i, j, and a third model k as arbitrary chosen baseline, and
    let's call p(i>j) is probability that i wins j:
      elo_score(i) = win_odd(i,k) * 2
      elo_score(k) = 1
      win_odd(i,j) = win_odd(i,k) * win_odd(k,j)  -> this is required to be true if score exists
      win_odd(i,j) = p(i>j) / (1-p(i>j))
      p(i>j) = elo_score(i) / (elo_score(i) + elo_score(j))

    Reference of the algorithm
    https://en.wikipedia.org/wiki/Bradley%E2%80%93Terry_model
    https://jmlr.org/papers/volume24/22-1086/22-1086.pdf

    Different algorithms converges at different speed. The algorithm used here is proposed in 2023
    in the paper referenced. It is 100x faster than previous Zermelo. The implementation below is also
    efficient with sparse matrix.
    """
    new_elo_scores = initial_elo_scores
    rating_error = elo_rating_error_bar
    for _ in range(100):
        prev_elo_scores = new_elo_scores.copy()
        for player in new_elo_scores.keys():
            new_elo_scores[player] = _get_updated_elo_score(new_elo_scores, player, wins_dict)
        rating_error = _get_elo_rating_error_bar(new_elo_scores, prev_elo_scores)
        if rating_error < elo_rating_error_bar:
            break
    if rating_error >= elo_rating_error_bar:
        raise ValueError('elo failed to converage with error={rating_error}')
    new_elo_scores = _normalize_elo_scores(new_elo_scores)
    return new_elo_scores


def _get_updated_elo_score(new_elo_scores, competitor_id, wins_dict):
    numerator = 0
    denominator = 0
    for opponent_id in wins_dict.get(competitor_id, {}).keys():
        score_sum = new_elo_scores[competitor_id] + new_elo_scores[opponent_id]
        numerator += wins_dict[competitor_id][opponent_id] * new_elo_scores[opponent_id] / score_sum
    for opponent_id in wins_dict.keys():
        score_sum = new_elo_scores[competitor_id] + new_elo_scores[opponent_id]
        denominator += wins_dict[opponent_id].get(competitor_id, 0) / score_sum
    score = LOWEST_ELO_SCORE
    if numerator > 0:
        score = numerator / denominator if denominator > 0 else HIGHEST_ELO_SCORE
    return score


def _get_elo_rating_error_bar(new_elo_scores, elo_scores):
    rating_error = 0
    if len(new_elo_scores):
        elo_score_diffs = [new_elo_scores[player] - elo_scores[player] for player in new_elo_scores.keys()]
        max_elo_score_diff = max(elo_score_diffs)
        min_finite_elo_scores = min([value for value in new_elo_scores.values() if value > LOWEST_ELO_SCORE])
        rating_error = max_elo_score_diff / min_finite_elo_scores * MAX_DELTA_ELO_RATING_PER_DELTA_SCORE_RATIO
    return rating_error


def _normalize_elo_scores(scores_dict):
    scores_to_average = [
        score for score in scores_dict.values()
        if 0 < score < HIGHEST_ELO_SCORE
    ]
    normalized = scores_dict
    if len(scores_to_average):
        average_elo_score = sum(scores_to_average) / len(scores_to_average)
        normalized = {
            player: score / average_elo_score if LOWEST_ELO_SCORE < score < HIGHEST_ELO_SCORE else score
            for player, score in scores_dict.items()
        }
    return normalized


def get_elo_ratings(scores, elo_base_score, elo_base_rating):
    elo_ratings = {
        player: get_elo_rating(score, elo_base_score, elo_base_rating)
        for player, score in scores.items()
    }
    return elo_ratings


def get_elo_rating(score, elo_base_score, elo_base_rating):
    if score <= LOWEST_ELO_SCORE:
        rating = LOWEST_ELO_RATING
    elif score >= HIGHEST_ELO_SCORE:
        rating = HIGHEST_ELO_RATING
    else:
        rating = 400 * log10(score / elo_base_score) + elo_base_rating
        rating = max(rating, LOWEST_ELO_RATING)
        rating = min(rating, HIGHEST_ELO_RATING)
        # different environment has different precision, and round to 2 digits is well
        # beyond necessary precision without incur mismatch in test
        rating = round(rating, 2)
    return rating



from datetime import datetime

from pydantic import BaseModel, validator
from typing import List

from chaiverse.config import FEED_ENDPOINT
from chaiverse.http_client import SubmitterClient
from chaiverse.lib.now import utcnow


class FeedEntry(BaseModel):
    event_type: str
    timestamp: str
    resource_id: str
    message: str
    user: dict

    @validator("timestamp", pre=True)
    def convert_timestamp(cls, timestamp):
        if isinstance(timestamp, datetime):
            timestamp = int(timestamp.timestamp())
        return str(timestamp)


class Feed(BaseModel):
    entries: List[FeedEntry]


def get_feed(limit, end_time=None):
    if end_time is None:
        end_time = int(utcnow().timestamp())
    client = SubmitterClient()
    params = {"end_time": end_time, "limit": limit}
    feed = client.get(endpoint=FEED_ENDPOINT, params=params)
    feed = Feed(entries=feed)
    return feed



import os


DEFAULT_MAX_WORKERS = 1


PUBLIC_LEADERBOARD_MINIMUM_FEEDBACK_COUNT = 0


SUBMISSION_CONSTANT_FIELD_DEFAULTS = {
    'double_thumbs_up': 0,
    'thumbs_up': 0,
    'thumbs_down': 0,
    'stay_in_character': float('nan'),
    'user_preference': float('nan'),
    'entertaining': float('nan'),
    'is_custom_reward': False,
    'safety_score': float('nan'),
    'elo_rating': float('nan'),
    'num_battles': 0,
    'num_wins': 0,
    'total_feedback_count': 0,
    'model_num_parameters': float('nan'),
}


LEADERBOARD_INDIVIDUAL_RANK_PARAMS = [
    dict(value_column='stay_in_character', rank_column='stay_in_character_rank', ascending=False),
    dict(value_column='user_preference', rank_column='user_preference_rank', ascending=False),
    dict(value_column='entertaining', rank_column='entertaining_rank', ascending=False),
    dict(value_column='double_thumbs_up_ratio', rank_column='double_thumbs_up_rank', ascending=False),
    dict(value_column='elo_rating', rank_column='elo_rank', ascending=False),
]


LEADERBOARD_OVERALL_RANK_PARAMS = [
    dict(
        from_rank_columns=['stay_in_character_rank', 'user_preference_rank', 'entertaining_rank'],
        overall_score_column='overall_meval_score', overall_rank_column='overall_meval_rank'
    ),
    dict(
        from_rank_columns=['elo_rank', 'double_thumbs_up_rank'],
        overall_score_column='overall_user_score', overall_rank_column='overall_user_rank'
    ),
]


LEADERBOARD_FORMAT_CONFIGS = {}


LEADERBOARD_FORMAT_CONFIGS['meval'] = {
    "output_columns": [
        'developer_uid',
        'model_name',
        'is_custom_reward',
        'stay_in_character',
        'user_preference',
        'entertaining',
        'safety_score',
        "overall_meval_score",
        'elo_rating',
        'num_battles',
        'num_wins',
        'thumbs_up_ratio',
        'double_thumbs_up_ratio',
        'size',
        'status',
        'submission_id',
    ],
    "sort_params": {
        "by": "overall_meval_score",
        "ascending": True
    },
}


LEADERBOARD_FORMAT_CONFIGS['thumbs_up'] = {
    "output_columns": [
        'developer_uid',
        'model_name',
        'double_thumbs_up_ratio',
        'elo_rating',
        'num_battles',
        'num_wins',
        'thumbs_up_ratio',
        'double_thumbs_up',
        'thumbs_up',
        'thumbs_down',
        'total_feedback_count',
        'overall_user_rank',
        'stay_in_character',
        'user_preference',
        'entertaining',
        'safety_score',
        'repetition',
        'is_custom_reward',
        'submission_id',
        'size',
    ],
    "sort_params": {
        "by": "double_thumbs_up_ratio",
        "ascending": False
    }
}

LEADERBOARD_FORMAT_CONFIGS['user'] = {
    "output_columns": [
        'developer_uid',
        'model_name',
        'double_thumbs_up_ratio',
        'elo_rating',
        'double_thumbs_up_rank',
        'elo_rank',
        'overall_user_score',
        'num_battles',
        'num_wins',
        'thumbs_up_ratio',
        'double_thumbs_up',
        'thumbs_up',
        'total_feedback_count',
        'stay_in_character',
        'user_preference',
        'entertaining',
        'safety_score',
        'is_custom_reward',
        'submission_id',
        'size',
    ],
    "sort_params": {
        "by": "overall_user_score",
        "ascending": True
    }
}



from datetime import datetime
from typing import Dict, Optional, List

import numpy as np
import pandas as pd
from pydantic import BaseModel, Extra, Field, root_validator

from chaiverse import config
from chaiverse.lib.now import utcnow
from chaiverse.schemas.leaderboard_row_schema import LeaderboardRow


DEFAULT_LEADERBOARD_HEADER_MAPPING = {
    'double_thumbs_up_ratio': '👍👍_ratio',
    'thumbs_up_ratio': '👍👍+👍_ratio',
    'single_thumbs_up_ratio': '👍_ratio',
    'thumbs_down_ratio': '👎_ratio',
}

DEFAULT_LEADERBOARD_INCLUDES = [
    'developer_uid',
    'submission_id',
    'celo_rating',
    'win_ratio',
    'num_battles',
    'double_thumbs_up_ratio',
    'thumbs_down_ratio',
    'feedback_count',
    'model_score',
    'safety_score',
    'best_of',
    'max_input_tokens',
    'model_repo',
    'reward_repo',
    'status',
    'model_name',
    'us_pacific_date',
]

DEFAULT_LEADERBOARD_EXCLUDES = []

DEFAULT_LEADERBOARD_SORT_PARAMS = {
    'by': 'celo_rating',
    'ascending': False
}

DEFAULT_TABULATE_OPTIIONS = {
    'numalign': 'decimal',
}


class Leaderboard(BaseModel, extra=Extra.allow):
    '''
    We are migrating from storing in
        leaderboard_rows of List[LeaderboardRow] to
        leaderboard_dict of Dict[submission_id, LeaderboardRow]
    During the transition:
        Leaderboard loaded from firebase can either contain list or dict.
        The validator will auto populat the leaderboard_dict if only leaderboard_rows existed.
        All code should read from leaderboard_dict and write to leaderboard_dict.
    After transition:
        The leaderboard_rows will be deleted from the code together with the validator.
        The old leaderboards stored in firebase will be deleted in data migration.
    '''
    timestamp: datetime = Field(default_factory=lambda: utcnow())
    leaderboard_rows: Optional[List[LeaderboardRow]]
    leaderboard_dict: Dict[str, LeaderboardRow]

    @root_validator(pre=True)
    def validate_leaderboard(cls, values):
        if not isinstance(values.get('leaderboard_dict'), dict):
            rows = values.get('leaderboard_rows', [])
            leaderboard_dict = {getattr(row, 'submission_id', None) or row['submission_id']: row for row in rows}
            values['leaderboard_rows'] = None
            values['leaderboard_dict'] = leaderboard_dict
        return values

    @property
    def rows(self) -> List[LeaderboardRow]:
        return self.leaderboard_dict.values()

    @property
    def df(self) -> pd.DataFrame:
        leaderboard_rows = [row.all_fields_dict() for row in self.rows]
        df = pd.DataFrame.from_records(leaderboard_rows)
        df = _sort_by_values(df, **DEFAULT_LEADERBOARD_SORT_PARAMS)
        if len(df) > 0:
            df = _add_repo_aggregate_columns(df)
        return df

    def to_display_df(self, includes=None, excludes=None, sort_params=None, header_mapping=None) -> pd.DataFrame:
        includes = includes or DEFAULT_LEADERBOARD_INCLUDES
        excludes = excludes or DEFAULT_LEADERBOARD_EXCLUDES
        sort_params = sort_params or DEFAULT_LEADERBOARD_SORT_PARAMS
        header_mapping = header_mapping or DEFAULT_LEADERBOARD_HEADER_MAPPING
        df = self.df
        if len(df) > 0:
            df = _include_listed_columns(df, includes)
            df = _exclude_listed_columns(df, excludes)
            df = _sort_by_values(df, **sort_params)
            df = df.round(2)
            df = df.rename(columns=header_mapping)
        return df

    def to_html(self, includes=None, header_mapping=None, sort_params=None):
        includes = includes or DEFAULT_LEADERBOARD_INCLUDES
        sort_params = sort_params or {"by": "celo_rating", "ascending": False}
        df = self.to_display_df(includes=includes, header_mapping=header_mapping, sort_params=sort_params)
        df.fillna("", inplace=True)
        html = df.to_html(classes="leaderboard-table table display nowrap", index=False, justify="center", escape=False)
        return html

    @property
    def auto_deactivation_candidates(self):
        rows = _get_can_auto_deactivate_rows(self.rows)
        rows = _sort_auto_deactivate_rows(rows)
        rows = _remove_top_models(rows)
        submission_ids = [row.submission_id for row in rows]
        return submission_ids

    def get_row(self, submission_id):
        df = self.df[self.df["submission_id"] == submission_id]
        row = {} if df.empty else df.to_dict(orient='records')[0]
        return row


def _add_repo_aggregate_columns(df):
    df['model_repo'] = df.get('model_repo', '')
    df['best_of'] = df.get('best_of', 1)
    for column in ["celo_rating", "double_thumbs_up_ratio", "thumbs_down_ratio", "win_ratio", "safety_score", "model_score"]:
        df[f"{column}_by_repo"] = df.groupby("model_repo")[column].transform("max")
    for column in ["feedback_count", "num_battles"]:
        df[f"{column}_by_repo"] = df.groupby("model_repo")[column].transform("sum")
    for column in ["celo_rating", "double_thumbs_up_ratio", "thumbs_down_ratio", "win_ratio", "safety_score", "model_score"]:
        df[f"{column}_by_repo_open"] = df.where(df["best_of"] == 1).groupby("model_repo")[column].transform("max")
    for column in ["feedback_count", "num_battles"]:
        df[f"{column}_by_repo_open"] = df.where(df["best_of"] == 1).groupby("model_repo")[column].transform("sum")
    return df


def _include_listed_columns(df, includes):
    df = df[[column for column in includes if column in df.columns]]
    return df


def _exclude_listed_columns(df, excludes):
    df = df[[column for column in df.columns if column not in excludes]]
    return df


def _sort_by_values(df, by: List[str], ascending: bool):
    if len(df) > 0:
        df = df.sort_values(by=by, ascending=ascending, na_position='last', ignore_index=True)
        df.index = np.arange(1, len(df) + 1)
    return df


def _get_can_auto_deactivate_rows(rows: List[LeaderboardRow]):
    rows = [row for row in rows if row.can_auto_deactivate()]
    return rows


def _sort_auto_deactivate_rows(rows: List[LeaderboardRow]):
    rows = sorted(rows, key=lambda row: row.celo_rating, reverse=True)
    return rows


def _remove_top_models(rows: List[LeaderboardRow]):
    rows = rows[config.AUTO_DEACTIVATION_MIN_RANK - 1:]
    return rows
